/*
** toybox.c
** Copyright (C) 2023  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
**
*****************************************************************************
**
** toybox: container exploration toy
**
** This program implements a primitive "container"
** under the Linux 4.4 kernel. It requires an
** existing target directory (which will become
** the container's root directory), populated with
** (at a minimum) the bin directory and a bin/sh
** binary. toybox will create any other directories
** that it requires for it's operation.
**
** Usage: toybox [-h hostname] rootdir
** 
** Where: -h hostname sets the container hostname to specified value
**
** toybox directory structure:
**
**   /
**   :
**   +- /bin		required, prepopulated, must include /bin/sh
**   :
**   +- /dev		constructed, toybox will bind mount host /dev to this mountpoint
**   :
**   +- /etc		optional, may hold prepopulated config files
**   :
**   +- /mnt		constructed, toybox will hold temp mounts here
**   :  :
**   :  +- /mnt/.proc	constructed, used by toybox to temporarily mount host /proc
**   :  :
**   :  +- /mnt/.root	constructed, used by toybox to temporarily mount host /
**   :  :
**   :  +- /mnt/.sys	constructed, used by toybox to temporarily mount host /sys
**   :
**   +- /proc		constructed, toybox will mount procfs to this mountpoint
**   :
**   +- /sbin		optional, may hold prepopulated binaries
**   :
**   +- /sys		constructed, toybox will mount sysfs to this mountpoint
**   :
**   +- /tmp		constructed, toybox will mount tmpfs to this mountpoint
**   :
**   +- /usr		optional, may hold prepopulated directories, binaries
**   :
**   +- /var		optional, may hold prepopulated data
**
**
** NB: compile with -D_GNU_SOURCE
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>

/* pivot_root() function prototype not supplied by standard headers */
extern int pivot_root(const char *new_root, const char *put_old);

#define	CONTAINER	(CLONE_NEWUSER	|	\
			 CLONE_NEWNS	|	\
			 CLONE_NEWPID	|	\
			 CLONE_NEWUTS	|	\
			 CLONE_NEWNET	|	\
			 CLONE_NEWIPC)
/*
** NB:	CLONE_NEWUSER	gives us a new user namespace, and the process
**			gains CAP_SYS_ADMIN
**	CLONE_NEWNS	gives us a new mount namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWPID	gives us a new PID namespace, which
**			gives access to the /proc filesystem,
**			and the special properties of PID 1
**			this namespace also requires the manipulating
**			process to have CAP_SYS_ADMIN
**	CLONE_NEWUTS	gives us a new UTS namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWNET	gives us a new NETWORK namespace, which
**			gives us access to an isolated set of network
**			resources. Requires CAP_SYS_ADMIN
**	CLONE_NEWIPC	gives us a new IPC namespace, which
**			gives us access to an isolated set of Sys V IPC
**			and POSIX message queue resources. Requires
**			CAP_SYS_ADMIN
**	CLONE_NEWCGROUP	is not implemented in kernel 4.4
*/


#define	SYSERROR(x,y)	((fprintf(stderr,"%s: %s returned \"%s\"\n",pgmId,x,strerror(errno))),(y))

struct tParameters {
  unsigned int	tFlags;
  char		*tRootdir;
  char		*tHostname;
};
typedef struct tParameters toyParms_t;
#define	PARM_HOSTNAME		(1<<0)


static int	toybox(toyParms_t toyParms);
static int 	assumeRoot(uid_t oldUid, gid_t oldGid, toyParms_t toyParms);
static int 	isolateFs(toyParms_t toyParms);
static int	idleProcess(toyParms_t toyParms);
static int	initProcess(toyParms_t toyParms);

static char *pgmId = "toybox";

int main(int argc, char *argv[])
{
  int	success = 1,		/* assume success */
	opt;
  toyParms_t	toyParameters = {0,NULL};

  if ((argc > 0) && argv[0] && *argv[0])	/* totally unnecessary safty check */
    pgmId = argv[0];

  while ((opt = getopt(argc,argv,"h:")) != -1)
  {
    switch (opt)
    {
      case 'h':
	toyParameters.tFlags |= PARM_HOSTNAME;
	free(toyParameters.tHostname);
	toyParameters.tHostname = strdup(optarg);
	break;

      default:
	success = 0;
	break;
    }
  }

  if (argc - optind != 1) success = 0;		/* wrong number of arguments */

  if (success)
  {
    toyParameters.tRootdir = argv[optind];
    success = toybox(toyParameters);
  }
  else
    fprintf(stderr,"Usage: %s [-h hostname] rootdir\n",argv[0]);

  return success?EXIT_SUCCESS:EXIT_FAILURE;
}

/*
** toybox(toyParms_t toyParms)
**
** Create and execute our toy container by building
** unshare()ing its namespaces, changing its uid and gid,
** building its root filesystem and executing its idle
** and init processes.
**
** Accepts:	toyParms_t containing option
**		 flags and associated values
**
** Returns:	int 1 if successful in creating container
**		int 0 if container creation failed
*/
static int toybox(toyParms_t toyParms)
{
  int	success = 1;	/* assume success */
  uid_t	host_uid;
  gid_t	host_gid;

  umask(0);
  host_uid = getuid();	/* must do before unshare() as uid changes within unshare */
  host_gid = getgid();	/* must do before unshare() as gid changes within unshare */

  /* change our namespaces */
  if (unshare(CONTAINER))				
    success = SYSERROR("unshare()",0);

  /* change our uid, gid to 0,0 */
  if (success)
    success = assumeRoot(host_uid,host_gid, toyParms);

  /* set up our isolated root directory */
  if (success)
    success = isolateFs(toyParms);

  /* run our idle and init processes */
  if (success)
    success = idleProcess(toyParms);

  return success;
}

/*
** assumeRoot(oldUid, oldGid, toyParms_t toyParms)
**
** Update the process entry in /proc to map our 
** outside-the-box uid and gid to the inside-the-box
** root user uid and gid, matching the effects
** of our new CLONE_NEWUSER capabilities.
**
** Accepts:	uid_t oldUid - our outside-the-box UID,
**		gid_t oldGid - our outside-the-box GID
**		toyParms_t containing parameter flags
**		  and associated values
**
** Returns: 	int 1  - success
**	    	int 0  - failure
**
** Note:	Does not use option flags/values at this time
*/
static int assumeRoot(uid_t oldUid, gid_t oldGid, toyParms_t toyParms)
{
  int	success = 1,	/* assume success */ 
	fd;
  char buff[48];	/* big enough for our needs */


  /*
  ** Map our uid into the container's root uid (UID 0)
  */
  sprintf(buff,"0 %lu 1\n",(unsigned long)oldUid);
  if ((fd = open("/proc/self/uid_map",O_WRONLY)) >= 0)
  {
    write(fd,buff,strlen(buff));
    close(fd);
  }
  else success =  SYSERROR("open(\"/proc/self/uid_map\")",0);

  /*
  ** Some kernels require /proc/self/setgroups set to "deny"
  ** before we can remap our GID
  */
  if (success  && ((fd = open("/proc/self/setgroups",O_WRONLY)) >= 0))
  {
    write(fd,"deny",4);
    close(fd);
  }

  if (success)
  {
    /*
    ** Map our gid into the container's root gid (GID 0)
    */
    sprintf(buff,"0 %lu 1\n",(unsigned long)oldGid);
    if ((fd = open("/proc/self/gid_map",O_WRONLY)) >= 0)
    {
      write(fd,buff,strlen(buff));
      close(fd);
    }
    else success = SYSERROR("open(\"/proc/self/gid_map\")",0);
  }

  return success;
}

/*
** isolateFs(toyParms_t toyParms)
**
** Build our container's root filesystem by creating
** required subdirectories in target rootdir, then
** mount()ing various requisite resources on those
** subdirectories.
**
** Accepts:	toyParms_t containing parameter flags
**		 and associated values
**
** Returns:	int 1 if successful creation of container
**		 root filesystem hierarchy
**		int 0 if container root filesystem cannot
**		 be properly built
**
** Note:	Does not use option flags/values at this time
**		This function will generate warning messages
**		 under certain circumstances. These informational
**		 messages can be ignored.
*/
static int isolateFs(toyParms_t toyParms)
{
  int	success = 1;	/* assume success */

  /* Ensure that none of our actions here affect host mountspace */
  if (mount("none","/",NULL,MS_REC|MS_PRIVATE,NULL))
    success = SYSERROR("mount(,\"/\",,)",0);

  /* make our root directory into a mountpoint in prep for pivot_root */
  if (success && mount(toyParms.tRootdir,toyParms.tRootdir,NULL,MS_BIND|MS_NOSUID,NULL))
    success = SYSERROR("mount(,faux_root,,)",0);

  /* Continue operations from within our intended root directory */
  if (success && chdir(toyParms.tRootdir))
    success = SYSERROR("chdir()",0);

  /* Remove and recreate any of the unpopulated, required subdirectories */
  if (success)
  {
    if (unlink("dev") && rmdir("dev"))			/* remove any residual /dev directory */
      SYSERROR("rmdir(\"dev\")",0);

    if (unlink("mnt/.proc") && rmdir("mnt/.proc"))	/* remove any residual /mnt/.proc directory */
      SYSERROR("rmdir(\"mnt/.proc\")",0);

    if (unlink("mnt/.root") && rmdir("mnt/.root"))	/* remove any residual /mnt/.root directory */
      SYSERROR("rmdir(\"mnt/.root\")",0);

    if (unlink("mnt/.sys") && rmdir("mnt/.sys"))	/* remove any residual /mnt/.sys directory */
      SYSERROR("rmdir(\"mnt/.sys\")",0);

    if (unlink("mnt") && rmdir("mnt"))			/* remove any residual /mnt directory */
      SYSERROR("rmdir(\"mnt\")",0);

    if (unlink("proc") && rmdir("proc"))		/* remove any residual /proc directory */
      SYSERROR("rmdir(\"proc\")",0);

    if (unlink("sys") && rmdir("sys"))			/* remove any residual /sys directory */
      SYSERROR("rmdir(\"sys\")",0);

    if (unlink("tmp") && rmdir("tmp"))			/* remove any residual /tmp directory */
      SYSERROR("rmdir(\"tmp\")",0);
  }

  if (success && mkdir("dev",0755))			/* create a new, empty /dev directory */
    success = SYSERROR("mkdir(\"dev\")",0);		/* to install our devices in */

  if (success && mkdir("mnt",0755))			/* create a new, empty /mnt directory */
    success = SYSERROR("mkdir(\"mnt\")",0);		/* to host some temporary mounts */

  if (success && mkdir("mnt/.proc",0755))		/* create a new, empty /mnt/.proc directory */
    success = SYSERROR("mkdir(\"mnt/.proc\")",0);	/* to temporarily mount a procfs to 	    */

  if (success && mkdir("mnt/.root",0751))		/* create a new, empty /mnt/.root directory */
    success = SYSERROR("mkdir(\"mnt/.root\")",0);	/* to temporarily mount our old root dir to */

  if (success && mkdir("mnt/.sys",0751))		/* create a new, empty /mnt/.sys directory */
    success = SYSERROR("mkdir(\"mnt/.sys\")",0);	/* to temporarily mount our old /sys dir to */

  if (success && mkdir("proc",0755))			/* create a new, empty /proc directory */
    success = SYSERROR("mkdir(\"proc\")",0);		/* to mount our procfs to */

  if (success && mkdir("sys",0755))			/* create a new, empty /sys directory */
    success = SYSERROR("mkdir(\"sys\")",0);		/* to mount our sysfs to */

  if (success && mkdir("tmp",01777))			/* create a new, empty /tmp directory */
    success = SYSERROR("mkdir(\"tmp\")",0);		/* to mount a tmpfs to */


  /* Populate our /dev directory with direct link to host /dev */
  if (success && mount("/dev","dev",NULL,MS_BIND|MS_REC,NULL))
    success = SYSERROR("mount(\"dev\")",0);

  /* Populate our (temporary) proc with direct link to host /proc */
  if (success && mount("/proc","mnt/.proc",NULL,MS_BIND|MS_REC,NULL))
    success = SYSERROR("mount(\"mnt/.proc\")",0);

  /* Populate our /sys directory with direct link to host /sys */
  if (success && mount("/sys","mnt/.sys",NULL,MS_BIND|MS_REC,NULL))
    success = SYSERROR("mount(\"mnt/.sys\")",0);

  /* Give ourselves a tempfs for our /tmp directory */
  if (success && mount("none","tmp","tmpfs",
			MS_NOSUID|MS_NODEV|MS_NOEXEC|MS_NOATIME,
			"size=16m,nr_inodes=4k,mode=770"))
    success = SYSERROR("mount(\"tmp\")",0);

  /* Now, Isolate our directory space and occupy the toy box */
  if (success && pivot_root(".","mnt/.root"))
    success = SYSERROR("pivot_root(\".\",\"mnt/.root\")",0);

  /* pivot_root(2) recommends that we chdir("/") to ensure correct cwd */
  if (success && chdir("/"))
    success = SYSERROR("chdir(\"/\")",0);

  /* Discard our tie to the host root directory */
  if (success && umount2("/mnt/.root",MNT_DETACH))
    success = SYSERROR("umount2(\"/mnt/.root\")",0);

  /* Discard the temporary directory on which we saved host root */
  if (success && unlink("/mnt/.root") && rmdir("/mnt/.root"))
    SYSERROR("rmdir(\"/mnt/.root\")",0);

  return success;
}

/*
** idleProcess(toyParms_t toyParms)
**
** Execute container's "idle" (PID 0) 
** and "init" (PID 1) processes
**
** Accepts:	toyParms_t containing parameter flags
**		 and associated values
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
*/
static int idleProcess(toyParms_t toyParms)
{
  int	success = 1,		/* assume success */
	child_status;
  pid_t	child;

  switch (child = fork())
  {
    case -1:	/* still in parent: fork() failed */
      success = SYSERROR("fork()",0);
      break;

    case 0:	/* in child, which takes the init process role */
      success = initProcess(toyParms);
      break;

    default:	/* in parent, which takes on the idle process role */
      wait(&child_status);
      if (!WIFEXITED(child_status) || (WEXITSTATUS(child_status) != 0))
	success = 0;

      /* unmount our private copies of mounted filesystems */
      umount2("/tmp",MNT_DETACH);	/* not strictly necessary */
      umount2("/dev",MNT_DETACH);	/* not strictly necessary */
      umount2("/sys",MNT_DETACH);	/* not strictly necessary */
      umount2("/proc",MNT_DETACH);	/* not strictly necessary */
  }

  return success;
}

/*
** initProcess(toyParms_t toyParms)
**
** Execute the container's init (PID 1) process
**
** Accepts:	toyParms_t containing parameter flags
**		 and associated values
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
**
** Note:	This function will generate warning messages
**		 under certain circumstances. These informational
**		 messages can be ignored.
*/
static int initProcess(toyParms_t toyParms)
{
  int	success = 1;		/* assume success */
  char	*envp[] = {"PATH=/sbin:/bin:/usr/sbin:/usr/bin:",NULL};

  /*
  ** >>>>>>>>>> Mount /proc and discard /mnt/.proc <<<<<<<<<<
  */
  /* Populate our /proc directory */
  if (mount("none","/proc","proc",MS_REC,NULL))
    success = SYSERROR("mount(\"/proc\")",0);

  /* Discard our tie to the host /proc directory */
  if (success && umount2("/mnt/.proc",MNT_DETACH))
    success = SYSERROR("umount2(\"/mnt/.proc\")",0);

  /* Discard the temporary directory on which we saved host proc */
  if (success && unlink("/mnt/.proc") && rmdir("/mnt/.proc"))
    SYSERROR("rmdir(\"mnt/.proc\")",0);


  /*
  ** >>>>>>>>>> Mount /sys and discard /mnt/.sys <<<<<<<<<<
  */
  /* Mount a true sysfs to our (isolated) /sys directory */
  if (success && mount("none","/sys","sysfs",0,NULL))
    success = SYSERROR("mount(\"/sys\")",0);

  /* Discard our tie to the host /sys directory */
  if (success && umount2("/mnt/.sys",MNT_DETACH))
    success = SYSERROR("umount2(\"/mnt/.sys\")",0);

  /* Discard the temporary directory on which we saved host /sys */
  if (success && unlink("/mnt/.sys") && rmdir("/mnt/.sys"))
    SYSERROR("rmdir(\"/mnt/.sys\")",0);


  /*
  ** >>>>>>>>>> Set hostname, run init process (shell) <<<<<<<<<<
  */
  /* If required, set our container's UTS hostname */
  if (success && (toyParms.tFlags & PARM_HOSTNAME))
    if (sethostname(toyParms.tHostname,strlen(toyParms.tHostname)))
      SYSERROR("sethostname()",0);

  /* Now, run a shell as our init process */
  if (success && execle("/bin/sh","/bin/sh",NULL,envp))
       success = SYSERROR("execle(\"/bin/sh\")",0);

  return success;
}

/*
** $Log: toybox.c,v $
** Revision 1.22  2023/01/27 01:59:47  lpitcher
** reverted code that creates own dev entries.
** will continue with valid code until we determine
** what caused the toydev failures
** This version logically identical to 1.20
**
** Revision 1.20  2023/01/22 14:30:34  lpitcher
** pass name of root directory in toyopts_t toyOptions structure
** remove rootdir reference from function declarations and definitions
** in recognition of expanded function of tOptions control block,
**   renamed tOptions to tParameters
**   renamed toyOpts_t to toyParms_t
**   renamed toyOpts to toyParms
**   renamed toyOptions to toyParameters
**   renamed FLAG_HOSTNAME to PARM_HOSTNAME
**
** Revision 1.19  2023/01/20 21:38:15  lpitcher
** after pivot_root(), chdir("/") to ensure correct cwd
** at idle termination, unmount our private copies of mounted filesystems
** pass toyopts_t toyParameters by value to subordinate functions
**
** Revision 1.18  2023/01/16 00:03:48  lpitcher
** added more commentary around directory manipulation
** isolate and mount our own sysfs to /sys
**
** Revision 1.17  2023/01/09 16:05:08  lpitcher
** use absolute paths once we pivot_root
**
** Revision 1.16  2023/01/09 15:44:40  lpitcher
** final cleanup of unused vars - fixed some unintended compile errors
**
** Revision 1.15  2023/01/09 15:43:44  lpitcher
** clean up unused vars
**
** Revision 1.14  2023/01/09 15:42:20  lpitcher
** simplified all isolateFs() tests
**
** Revision 1.13  2023/01/09 15:35:43  lpitcher
** simplified all initProcess() tests
**
** Revision 1.12  2023/01/09 15:31:37  lpitcher
** simplified sethostname call
**
** Revision 1.11  2023/01/09 15:29:30  lpitcher
** adjusted init PATH
**
** Revision 1.10  2023/01/09 13:31:46  lpitcher
** expanded container creation commentary
**
** Revision 1.9  2023/01/09 13:21:16  lpitcher
** replaced flags and hostname with toyParms_t structure
** set flag in toyParms_t for hostname when given
** use flag in toyParms_t to determine hostname when needed
**
** Revision 1.8  2023/01/09 07:29:03  lpitcher
** allow empty hostname to reset UTS hostname
**
** Revision 1.7  2023/01/09 07:00:21  lpitcher
** correct spelling errors
**
** Revision 1.6  2023/01/09 06:42:23  lpitcher
** propogate flags variable to assumeRoot(), isolateFs(), idleProcess(), and initProcess()
** global rename status variable to success, to better reflect its purpose
**
** Revision 1.5  2023/01/09 06:02:47  lpitcher
** refactored toybox() logic
** added parens to SYSERROR macro
**
** Revision 1.4  2023/01/09 05:43:19  lpitcher
** added empty hostname check to initProcess
** added descriptive comments for each function
**
** Revision 1.3  2023/01/09 05:30:53  lpitcher
** added sethostname() to initProcess()
** added hostname parameter to idleProcess(), initProcess()
** pass hostname to idleProcess, initProcess
**
** Revision 1.2  2023/01/09 05:18:46  lpitcher
** added RCS log
**
*/
